/*
 *  $Id: packet.c,v 1.2 1998/10/19 23:36:39 route Exp $
 *
 *  Firewalk
 *  packet implementation
 *  packet.c - Packet verification and printing routines
 *
 *  Copyright (c) 1998 Mike D. Schiffman <mds@es2.net>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "./firewalk.h"
#include "./packet.h"


int
icmp_verify(u_char *packet, struct firepack *fp)
{
    struct ip *ip_hdr;
    struct icmphdr *icmp_hdr;
    struct ip *origip_hdr;
    struct udphdr *origudp_hdr;

    ip_hdr   = (struct ip *)(packet + fp->packet_offset);
    icmp_hdr = (struct icmphdr *)(packet + fp->packet_offset + IP_H);

    switch (icmp_hdr->type)
    {
        case ICMP_DEST_UNREACH:
        case ICMP_TIME_EXCEEDED:
            /*
             *  The ICMP error message contains the IP header and first 8
             *  bytes of data of datagram that caused the error.
             */
            origip_hdr = (struct ip *)(packet + fp->packet_offset + IP_H + ICMP_H);

            /*
             *  Was this a UDP or TCP packet that caused the problem?  If not,
             *  forget it, it's not ours.
             */
            if (origip_hdr->ip_p != fp->protocol) break;

            /*
             *  We can fudge a bit here.  We only really care about the
             *  transport layer port numbers, not really the transport protocol
             *  itself.  Since TCP and UDP both keep source and destination
             *  port numbers in the first 4 bytes, we can cast the packet as
             *  having a UDP header.
             */
            origudp_hdr = (struct udphdr *)(packet + fp->packet_offset + 2 * IP_H + ICMP_H);

            /*
             *  We look for our packet ID and our source port.
             */
            if (ntohs(origip_hdr->ip_id) == fp->id &&
                ntohs(origudp_hdr->uh_sport) == fp->sport)
            {
                /*
                 *  This is a response from the gateway host.
                 */
                if (ip_hdr->ip_src.s_addr == fp->gateway) return (GW_REPLY);
                /*
                 *  This is a response from the destination host.
                 */
                if (icmp_hdr->type == ICMP_DEST_UNREACH &&
                    ip_hdr->ip_src.s_addr == fp->destination) return (DEST_REPLY);
            /*
             *  This is just a standard TTL expired reply.
             */
            return (REPLY);
            }
        default:
            break;
    }
    return (NO_REPLY);                 /* Not a firewalk packet */
}


int
process_response(u_short *cport, int response)
{
    switch (response)
    {
        case NO_REPLY:      /* packet was not for this process */
            return (0);
        case REPLY:         /* non-interesting host repsonse */
            fp->p_reply++;
            if (fp->red_flag)
            {
                if (!fp->quiet) fprintf(stdout, " .");
                break;
            }
            fp->red_flag = 1;
            if (!fp->quiet)
            {
                if (!fp->bound)
                {
                    fprintf(stdout, "<response from> ");
                    print_ip(packet, fp);
                }
                else
                {
                    fprintf(stdout, "open");
                }
            }
            else if (fp->bound) fprintf(stdout, "%d\n", *cport);
            break;
        case GW_REPLY:      /* gateway host response */
            fp->p_reply++;
            if (fp->bound && fp->red_cnt == 1)
            {
                fprintf(stdout, "Expired message from gateway...\n");
                return (-1);
            }
            if (fp->red_flag)
            {
                if (!fp->quiet) fprintf(stdout, " .");
                break;
            }
            fp->red_flag = 1;
            if (!(fp->gway_hop_cnt))
            {
                /*
                 *  We've hit the gateway, are bound, and can begin scanning.
                 */
                fp->bound = 1;
                fp->gway_hop_cnt = fp->ttl;
                if (!fp->quiet)
                {
                    fprintf(stdout, "Bound scan: %d hops", fp->gway_hop_cnt);
                }
            }
            if (!fp->quiet)
            {
                fprintf(stdout, " <Gateway at %d hops> ", fp->gway_hop_cnt);
                print_ip(packet, fp);
                fprintf(stdout, "\n");
            }
            break;
        case DEST_REPLY:        /* destination host response */
            fp->p_reply++;
            if (fp->red_flag)
            {
                fprintf(stdout, " .");
                break;
            }
            fp->red_flag = 1;
            fp->dest_hop_cnt = fp->ttl;
            if (!fp->quiet)
            {
                fprintf(stdout, "<Destination at %d hops> ", fp->dest_hop_cnt);
                print_ip(packet, fp);
            }
            if (!fp->bound)
            {
                /*
                 *  If we hit the destination and are not bound, the gateway
                 *  host does not lie between us and the destination host.
                 */
                fprintf(stdout,
                        "\nUnbound destination.  Gateway not en route to destination host.\n");
                return (-1);
            }
            break;
    }
    return (0);
}


void
print_ip(u_char *packet, struct firepack *fp)
{
    struct ip *ip_hdr;

    ip_hdr = (struct ip *)(packet + fp->packet_offset);
    fprintf(stdout, "[%s]", host_lookup(ip_hdr->ip_src.s_addr, fp->use_name));
}


void
p_stats(struct firepack *fp)
{
    fprintf(stdout,
            "%ld packets sent, %ld replies received\n",
            fp->p_sent,
            fp->p_reply);
}

/* EOF */
